from os.path import join, realpath

import fs

# Add the local code directory to the `fs` module path
# Can only rely on fs.__path__ being an iterable - on windows it's not a list
newPath = list(fs.__path__)
newPath.insert(0, realpath(join(__file__, '..', '..', 'fs')))
fs.__path__ = newPath



#!/usr/bin/env python

from json import dump, load
from logging import basicConfig, DEBUG
from os import environ
from sys import stdout

from pyperclip import copy
from requests_oauthlib import OAuth2Session

from github import UploadSecret

class TokenStorageFile:
	def __init__(self, path):
		self.path = path

	def Save(self, token_):
		with open(self.path, 'w', encoding='utf-8') as f:
			dump(token_, f)

	def Load(self):
		try:
			with open(self.path, encoding='utf-8') as f:
				return load(f)
		except FileNotFoundError:
			return None

def Authorize(clientId, clientSecret, redirectUri, storagePath):
	authorizationBaseUrl = 'https://login.microsoftonline.com/consumers/oauth2/v2.0/authorize'
	tokenUrl = 'https://login.microsoftonline.com/consumers/oauth2/v2.0/token'
	session = OAuth2Session(client_id=clientId, redirect_uri=redirectUri, scope=['offline_access', 'Files.ReadWrite'])
	authorizationUrl, _ = session.authorization_url(authorizationBaseUrl)
	print(f'Go to the following URL and authorize the app: {authorizationUrl}')

	copy(authorizationUrl)
	print('URL copied to clipboard')

	redirectResponse = input('Paste the full redirect URL here:')

	tokenStorage = TokenStorageFile(storagePath)

	environ['OAUTHLIB_RELAX_TOKEN_SCOPE'] = 'some value' # noqa: S105
	token_ = session.fetch_token(tokenUrl, client_secret=clientSecret, authorization_response=redirectResponse, include_client_id=True)
	tokenStorage.Save(token_)
	if 'XGITHUB_API_PERSONAL_TOKEN' in environ:
		UploadSecret(token_)
	return token_

def EscapeForBash(token_):
	charactersToEscape = '{}"[]: *!+/~^()'
	for character in charactersToEscape:
		token_ = token_.replace(character, '\\' + character)
	return token_

if __name__ == '__main__':
	basicConfig(stream=stdout, level=DEBUG, format='{levelname[0]}|{module}|{lineno}|{message}', style='{')
	token = Authorize(
		environ['GRAPH_API_CLIENT_ID'],
		environ['GRAPH_API_CLIENT_SECRET'],
		environ['GRAPH_API_REDIRECT_URI'],
		environ['GRAPH_API_TOKEN_PATH']
	)
	print(token)



from datetime import datetime, timedelta, timezone
from hashlib import sha1
from io import BytesIO
from json import dump, load, loads
from logging import info, warning
from os import environ
from time import sleep
from unittest import TestCase
from urllib.parse import parse_qs, urlencode, urlparse
from uuid import uuid4

from fs.onedrivefs import OneDriveFS, OneDriveFSOpener
from fs.opener import open_fs, registry
from fs.subfs import SubFS
from fs.test import FSTestCases
from ngrok import forward
from pytest import fixture, mark, raises
from pytest_localserver.http import WSGIServer

from .github import UploadSecret

_SAFE_TEST_DIR = 'Documents/test-onedrivefs'

class TokenStorageReadOnly:
	def __init__(self, token):
		self.token = token

	def Save(self, token):
		self.token = token
		UploadSecret(token)

	def Load(self):
		return loads(self.token)

class TokenStorageFile:
	def __init__(self, path):
		self.path = path

	def Save(self, token):
		with open(self.path, 'w', encoding='utf-8') as f:
			dump(token, f)

	def Load(self):
		try:
			with open(self.path, encoding='utf-8') as f:
				return load(f)
		except FileNotFoundError:
			return None

class SimpleApp:
	def __init__(self):
		self.notified = False

	def __call__(self, environ_, start_response):
		"""Simplest possible WSGI application"""
		status = '200 OK'
		response_headers = [('Content-type', 'text/plain')]
		start_response(status, response_headers)
		parsedQS = parse_qs(environ_['REQUEST_URI'][2:])
		info(f'Received: {parsedQS}')
		info(f'env: {environ_}')
		if 'validationToken' in parsedQS:
			info('Validating subscription')
			return [parsedQS['validationToken'][0].encode()]
		inputStream = environ_['wsgi.input']
		info(f'Input: {inputStream}')
		info('NOTIFIED')
		self.notified = True
		return ''

@fixture(scope='class')
def testserver(request):
	server = WSGIServer(application=SimpleApp())
	request.cls.server = server
	server.start()
	yield server
	server.stop()

def CredentialsStorage():
	if 'GRAPH_API_TOKEN_READONLY' in environ:
		return TokenStorageReadOnly(environ['GRAPH_API_TOKEN_READONLY'])
	return TokenStorageFile(environ['GRAPH_API_TOKEN_PATH'])

storage = CredentialsStorage() # keep at module level so that it can save and load credentials after refresh

def FullFS():
	return OneDriveFS(environ['GRAPH_API_CLIENT_ID'], environ['GRAPH_API_CLIENT_SECRET'], storage.Load(), storage.Save)

def test_list_root():
	fs = FullFS()
	assert fs.listdir('/') == fs.listdir('')

def test_opener_format():
	registry.install(OneDriveFSOpener())
	client_id = environ['GRAPH_API_CLIENT_ID']
	client_secret = environ['GRAPH_API_CLIENT_SECRET']
	credentials = storage.Load()
	access_token = credentials['access_token']
	refresh_token = credentials['refresh_token']

	encodedParameters = urlencode({'access_token': access_token, 'refresh_token': refresh_token, 'client_id': client_id, 'client_secret': client_secret})

	# Without the initial "/" character, it should still be assumed to relative to the root
	fs = open_fs(f'onedrive://{_SAFE_TEST_DIR}?' + encodedParameters)
	assert isinstance(fs, SubFS), str(fs)
	assert fs._sub_dir == f'/{_SAFE_TEST_DIR}' # noqa: SLF001

	# It should still accept the initial "/" character
	fs = open_fs(f'onedrive:///{_SAFE_TEST_DIR}?' + encodedParameters)
	assert isinstance(fs, SubFS), str(fs)
	assert fs._sub_dir == f'/{_SAFE_TEST_DIR}' # noqa: SLF001

class PyFsCompatLayer:
    """PyFilesystem2 Python 3.12 compatibility layer.

    Adds a workaround for PyFilesystem2#568:
    https://github.com/PyFilesystem/pyfilesystem2/issues/568
    """

    assertRaisesRegexp = TestCase.assertRaisesRegex

class TestOneDriveFS(FSTestCases, TestCase, PyFsCompatLayer):
	def make_fs(self):
		self.fullFS = FullFS()
		self.testSubdir = f'/{_SAFE_TEST_DIR}/{uuid4()}'
		return self.fullFS.makedirs(self.testSubdir)

	def destroy_fs(self, _):
		self.fullFS.removetree(self.testSubdir)

	@mark.skipif('NGROK_AUTHTOKEN' not in environ, reason='Missing NGROK_AUTHTOKEN environment variable')
	@mark.usefixtures('testserver')
	def test_subscriptions(self):
		port = urlparse(self.server.url).port
		info(f'Port: {port}')
		info(self.server.url)
		listener = forward(port, authtoken_from_env=True)
		info(f'publicUrl: {listener.url()}')
		expirationDateTime = datetime.now(timezone.utc) + timedelta(minutes=60)
		id_ = self.fs.create_subscription(listener.url(), expirationDateTime, 'client_state')
		info(f'subscription id: {id_}')
		self.fs.touch('touched-file.txt')
		info('Touched the file, waiting...')
		# need to wait for some time for the notification to come through, but also process incoming http requests
		for _ in range(50):
			if self.server.app.notified is True:
				break
			sleep(5)
		info('Sleep done, deleting subscription')
		self.fs.delete_subscription(id_)
		info('subscription deleted')
		assert self.server.app.notified is True, f'Not notified: {self.server.app.notified}'

	def test_overwrite_file(self):
		with self.fs.open('small_file_to_overwrite.bin', 'wb') as f:
			f.write(b'x' * 10)

		with self.fs.open('small_file_to_overwrite.bin', 'wb') as f:
			f.write(b'y' * 10)

		with self.fs.open('small_file_to_overwrite.txt', 'w') as f:
			f.write('x' * 10)

		with self.fs.open('small_file_to_overwrite.txt', 'w') as f:
			f.write('y' * 10)

		with self.fs.open('large_file_to_overwrite.bin', 'wb') as f:
			f.write(b'x' * 4000000)

		with self.fs.open('large_file_to_overwrite.bin', 'wb') as f:
			f.write(b'y' * 4000000)

		with self.fs.open('large_file_to_overwrite.txt', 'w') as f:
			f.write('x' * 4000000)

		with self.fs.open('large_file_to_overwrite.txt', 'w') as f:
			f.write('y' * 4000000)

	def test_photo_metadata(self):
		with self.fs.open('canon-ixus.jpg', 'wb') as target, open('tests/canon-ixus.jpg', 'rb') as source:
			target.write(source.read())

		# sometimes it take a few seconds for the server to process EXIF data
		# until it's processed, the "photo" section should be missing
		iterations = 50
		sleepTime = 5
		for _ in range(iterations):
			info_ = self.fs.getinfo('canon-ixus.jpg')

			self.assertTrue(info_.get('photo', 'camera_make') in {None, 'Canon'})
			self.assertTrue(info_.get('photo', 'camera_model') in {None, 'Canon DIGITAL IXUS'})
			self.assertTrue(info_.get('photo', 'exposure_denominator') in {None, 350})
			self.assertTrue(info_.get('photo', 'exposure_numerator') in {None, 1})
			self.assertTrue(info_.get('photo', 'focal_length') in {None, 10.8125})
			self.assertTrue(info_.get('photo', 'f_number') in {None, 4.0})
			self.assertTrue(info_.get('photo', 'taken_date_time') in {None, datetime(2001, 6, 9, 15, 17, 32)})
			self.assertTrue(info_.get('photo', 'iso') is None)
			self.assertTrue(info_.get('image', 'width') in {None, 640})
			self.assertTrue(info_.get('image', 'height') in {None, 480})
			if info_.get('photo', 'camera_make') is not None:
				break
			sleep(sleepTime)
		else:
			self.fail(f'EXIF metadata not processed in {iterations * sleepTime}s')

	def test_photo_metadata2(self):
		with self.fs.open('DSCN0010.jpg', 'wb') as target, open('tests/DSCN0010.jpg', 'rb') as source:
			target.write(source.read())

		# sometimes it take a few seconds for the server to process EXIF data
		# until it's processed, the "photo" section should be missing
		iterations = 50
		sleepTime = 5
		for iteration in range(iterations):
			info_ = self.fs.getinfo('DSCN0010.jpg')

			self.assertTrue(info_.get('photo', 'camera_make') in {None, 'NIKON'})
			self.assertTrue(info_.get('photo', 'camera_model') in {None, 'COOLPIX P6000'})
			self.assertTrue(info_.get('photo', 'exposure_denominator') in {None, 300.0})
			self.assertTrue(info_.get('photo', 'exposure_numerator') in {None, 4.0})
			self.assertTrue(info_.get('photo', 'focal_length') in {None, 24.0})
			self.assertTrue(info_.get('photo', 'f_number') in {None, 5.9})
			self.assertTrue(info_.get('photo', 'taken_date_time') in {None, datetime(2008, 10, 22, 16, 28, 39)})
			self.assertTrue(info_.get('photo', 'iso') in {None, 64})
			self.assertTrue(info_.get('image', 'width') in {None, 640})
			self.assertTrue(info_.get('image', 'height') in {None, 480})
			self.assertTrue(info_.get('location', 'latitude') in {None, 43.46744833333334})
			self.assertTrue(info_.get('location', 'longitude') in {None, 11.885126666663888})
			if info_.get('photo', 'camera_make') is not None:
				break
			warning(f'EXIF metadata not processed in {iteration * sleepTime}s')
			sleep(sleepTime)
		else:
			self.fail(f'EXIF metadata not processed in {iterations * sleepTime}s')

	def test_hashes(self):
		with self.fs.open('DSCN0010.jpg', 'wb') as target, open('tests/DSCN0010.jpg', 'rb') as source:
			data = source.read()
			target.write(data)

		hash_ = sha1() # noqa: S324
		hash_.update(data)

		# It takes time for the server to calculate the hashes
		iterations = 50
		sleepTime = 5
		for iteration in range(iterations):
			info_ = self.fs.getinfo('DSCN0010.jpg')
			if info_.get('hashes', 'SHA1') is not None:
				break
			warning(f'Hashes not calculated in {iteration * sleepTime}s')
			sleep(sleepTime)
		else:
			self.fail(f'Hashes not calculated in {iterations * sleepTime}s')
		self.assertEqual(hash_.hexdigest().upper(), info_.get('hashes', 'SHA1'))

	def test_download_as_format(self):
		with self.fs.open('a.md', 'w') as f:
			f.write('test')

		byteStream = BytesIO()
		self.fs.download_as_format('a.md', byteStream, 'html')
		byteStream.seek(0)
		data = byteStream.read()
		assert data.startswith(b'<p>'), data

		with open('tests/sample1.heic', 'rb') as f:
			self.fs.upload('sample1.heic', f)

		byteStream = BytesIO()
		self.fs.download_as_format('sample1.heic', byteStream, 'jpg', width=100, height=100)
		byteStream.seek(0)
		data = byteStream.read()
		assert data.startswith(b'\xFF\xD8\xFF'), data

		with raises(ValueError):
			self.fs.download_as_format('sample1.heic', BytesIO(), 'jpg')

		with raises(ValueError):
			self.fs.download_as_format('sample1.heic', BytesIO(), 'jpg', width=42.2, height=42)

		with raises(ValueError):
			self.fs.download_as_format('sample1.heic', BytesIO(), 'jpg', height=42)



# (generated with --quick)

import datetime as _datetime
import io
import logging
import requests.models
from typing import Annotated, Any, Callable, Dict, Mapping, Optional, Sequence, Tuple, Type, TypeVar, Union

BytesIO: Type[io.BytesIO]
DestinationExists: Any
DirectoryExists: Any
DirectoryExpected: Any
DirectoryNotEmpty: Any
FS: Any
FileExists: Any
FileExpected: Any
Info: Any
Mode: Any
OAuth2Session: Any
ResourceNotFound: Any
ResourceType: Any
SubFS: Any
_log: logging.Logger
basename: Any
datetime: Type[_datetime.datetime]
datetime_to_epoch: Any
dirname: Any
epoch_to_datetime: Any
timezone: Type[_datetime.timezone]

AnyStr = TypeVar('AnyStr', str, bytes)
_T2 = TypeVar('_T2')

class OneDriveFS(Any):
    _drive_root: str
    _meta: Dict[str, Optional[Union[bool, str]]]
    _resource_root: str
    _service_root: str
    session: OneDriveSession
    subfs_class: Type[SubOneDriveFS]
    def __init__(self, clientId, clientSecret, token, SaveToken, tenant = ..., **kwargs) -> None: ...
    def __repr__(self) -> str: ...
    def _itemInfo(self, item) -> Any: ...
    def copy(self, src_path, dst_path, overwrite = ..., preserve_time = ...) -> None: ...
    def create_subscription(self, notification_url, expiration_date_time, client_state) -> Any: ...
    def delete_subscription(self, id_) -> None: ...
    def download_as_format(self, path, output_file, format, **options) -> None: ...
    def getinfo(self, path, namespaces = ...) -> Any: ...
    def listdir(self, path) -> list: ...
    def makedir(self, path, permissions = ..., recreate = ...) -> Any: ...
    def move(self, src_path, dst_path, overwrite = ..., preserve_time = ...) -> None: ...
    def openbin(self, path, mode = ..., buffering = ..., **options) -> _UploadOnClose: ...
    def remove(self, path) -> None: ...
    def removedir(self, path) -> None: ...
    def scandir(self, path, namespaces = ..., page = ...) -> list: ...
    def set_drive(self, driveId = ..., userId = ..., groupId = ..., siteId = ...) -> None: ...
    def setinfo(self, path, info) -> None: ...
    def update_subscription(self, id_, expiration_date_time) -> None: ...

class OneDriveSession(Any):
    _drive_root: Any
    def __init__(self, *args, drive_root, **kwargs) -> None: ...
    def delete_item(self, path, extra = ..., **kwargs) -> Any: ...
    def delete_path(self, path, extra = ..., **kwargs) -> Any: ...
    def get_item(self, path, extra = ..., **kwargs) -> Any: ...
    def get_path(self, path, extra = ..., **kwargs) -> Any: ...
    def item_url(self, itemId, extra) -> str: ...
    def patch_item(self, path, extra = ..., **kwargs) -> Any: ...
    def path_url(self, path, extra) -> str: ...
    def post_item(self, path, extra = ..., **kwargs) -> Any: ...
    def post_path(self, path, extra = ..., **kwargs) -> Any: ...
    def put_item(self, path, extra = ..., **kwargs) -> Any: ...

class SubOneDriveFS(Any):
    def create_subscription(self, notification_url, expiration_date_time, client_state) -> Any: ...
    def delete_subscription(self, id_) -> Any: ...
    def download_as_format(self, path, output_file, format, **options) -> Any: ...
    def update_subscription(self, id_, expiration_date_time) -> Any: ...

class _UploadOnClose(io.BytesIO):
    _closed: bool
    closed: Annotated[bool, 'property']
    itemId: Any
    mode: Annotated[Any, 'property']
    parsedMode: Any
    path: Any
    session: Any
    def _ResumableUpload(self, itemId, filename) -> None: ...
    def __init__(self, session, path, itemId, mode) -> None: ...
    def close(self) -> None: ...
    def read(self, size = ...) -> bytes: ...
    def readable(self) -> Any: ...
    def truncate(self, size = ...) -> int: ...
    def writable(self) -> Any: ...
    def write(self, data) -> int: ...

def _FormatDateTime(dt) -> Any: ...
def _HandleError(response) -> None: ...
def _ParseDateTime(dt) -> _datetime.datetime: ...
def _UpdateDict(dict_, sourceKey, targetKey: _T2, processFn = ...) -> Dict[_T2, Any]: ...
def get(url: Union[bytes, str], params = ..., data = ..., headers = ..., cookies = ..., files = ..., auth = ..., timeout = ..., allow_redirects: bool = ..., proxies = ..., hooks = ..., stream = ..., verify = ..., cert = ..., json = ...) -> requests.models.Response: ...
def getLogger(name: Optional[str] = ...) -> logging.Logger: ...
def urlencode(query: Union[Mapping, Sequence[Tuple[Any, Any]]], doseq: bool = ..., safe: AnyStr = ..., encoding: str = ..., errors: str = ..., quote_via: Callable[[str, AnyStr, str, str], str] = ...) -> str: ...



from functools import wraps
from logging import getLogger
from time import sleep

from niquests import codes

_log = getLogger(__name__)

def throttle():
	def decorator(func):
		@wraps(func)
		def wrapper(*args, **kwargs):
			while True:
				resp = func(*args, **kwargs)
				if resp.status_code != codes.too_many_requests:
					break
				# look at the response and retry after a delay
				retryAfterSeconds = int(resp.headers['Retry-After'])
				_log.info(f'Sleeping for {retryAfterSeconds} sec after throttling')
				sleep(retryAfterSeconds)
			return resp
		return wrapper
	return decorator
